@@ -1,5 +1,16 @@
 # coding:utf-8
-from typing import Iterator, Union, Tuple, Mapping, List, Set, Callable, Optional as Opt
+from typing import (
+    Iterator,
+    Union,
+    Tuple,
+    Mapping,
+    List,
+    Set,
+    Callable,
+    Type,
+    Optional as Opt,
+)
+import copy
 import os
 import sys
 import shlex
@@ -21,8 +32,8 @@ from argparse import (
     SUPPRESS,
     _SubParsersAction,
 )
-from cytoolz import identity, get_in
-from typing_inspect import is_generic_type
+from traceback import print_exception
+from typing_inspect import is_generic_type, get_generic_type
 
 from bourbaki.introspection.classes import classpath, most_specific_constructor
 from bourbaki.introspection.types import (
@@ -31,7 +42,9 @@ from bourbaki.introspection.types import (
     get_param_dict,
     get_generic_origin,
     is_optional_type,
+    get_generic_args,
 )
+from bourbaki.introspection.generic_dispatch import GenericTypeLevelSingleDispatch
 from bourbaki.introspection.typechecking import isinstance_generic
 from bourbaki.introspection.docstrings import parse_docstring, CallableDocs
 
@@ -65,15 +78,19 @@ from .helpers import (
     _combined_cli_sig,
     _type,
     _validate_lookup_order,
+    identity,
     get_task,
     NamedChainMap,
     strip_command_prefix,
+    get_in,
     update_in,
     VARIADIC_NARGS,
 )
 from .decorators import cli_attrs, NO_OUTPUT_HANDLER
 from .signatures import CLISignatureSpec, FinalCLISignatureSpec
 
+__all__ = ["CommandLineInterface", "ArgSource", "DEFAULT_LOOKUP_ORDER"]
+
 # only need to parse docs once for any function
 parse_docstring = lru_cache(None)(parse_docstring)
 
@@ -95,9 +112,12 @@ RESERVED_NAMESPACE_ATTRS = (
 OUTPUT_GROUP_NAME = "output control"
 OPTIONAL_ARG_TEMPLATE = "{}??"
 MIN_VERBOSITY = 1
+TRACEBACK_VERBOSITY = 3
 DEFAULT_EXECUTION_FLAG = "-x"
 INSTALL_SHELL_COMPLETION_FLAG = "--install-bash-completion"
 CONFIG_OPTION = "--config"
+VERSION_FLAG = "--version"
+INFO_FLAG = "--info"
 VERBOSE_FLAGS = ("-v", "--verbose")
 QUIET_FLAGS = ("--quiet", "-q")
 EXECUTE = False
@@ -158,6 +178,45 @@ class WideHelpFormatter(RawDescriptionHelpFormatter):
         )
 
 
+# error handling
+
+
+class CLIErrorHandlingContext:
+    def __init__(
+        self, exit_codes: Mapping[Type[Exception], int] = None, verbose: bool = False
+    ):
+        if exit_codes is None:
+            exit_codes = {Exception: 1}
+        self.verbose = verbose
+        self.exit_codes = exit_codes
+        self.exit_code = GenericTypeLevelSingleDispatch("exit_code")
+        self.exit_code.register_from_mapping(exit_codes, as_const=True)
+
+    def __enter__(self):
+        pass
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        if exc_type is None or exc_val is None:
+            pass
+        else:
+            traceback = exc_tb if self.verbose else None
+            if exc_type is SystemExit and SystemExit not in self.exit_codes:
+                # exit with the same code when a SystemExit was raised elsewhere
+                sys.exit(exc_type.code)
+            else:
+                handlers = self.exit_code.all_resolved_funcs(exc_type)
+                if not handlers:
+                    # no matching exceptions; print full traceback and exit with the highest code
+                    code = 255
+                    traceback = exc_tb
+                else:
+                    # maximum code for the most specific exceptions registered
+                    code = max(handler(exc_type) for handler in handlers)
+
+            print_exception(exc_type, exc_val, traceback, chain=False, file=sys.stderr)
+            sys.exit(code)
+
+
 # The main CLI class and helpers
 
 
@@ -181,6 +240,7 @@ class _SubparserPathAction(_SubParsersAction):
 class PicklableArgumentParser(ArgumentParser):
     cmd_prefix = ()
 
+    # this shim makes argument parsers picklable (argparse argument parsers are not)
     def __init__(self, *args, **kwargs):
         kwargs["formatter_class"] = WideHelpFormatter
         super().__init__(*args, **kwargs)
@@ -188,6 +248,10 @@ class PicklableArgumentParser(ArgumentParser):
         self.register("type", None, identity)
         self.register("action", "parsers", _SubparserPathAction)
 
+    ##############
+    # subparsers #
+    ##############
+
     def add_subparsers(self, *args, **kwargs) -> _SubparserPathAction:
         subparsers = super().add_subparsers(*args, **kwargs)
         subparsers.cmd_prefix = self.cmd_prefix
@@ -207,7 +271,12 @@ class PicklableArgumentParser(ArgumentParser):
         self._subparsers_action = subparsers
         return subparsers
 
-    def get_nested_subparser(self, *cmd_path: str):
+    def get_nested_subparser(
+        self,
+        *cmd_path: str,
+        prefix: Tuple[str, ...] = (),
+        subcommand_help: Opt[Mapping[Tuple[str, ...], str]] = None,
+    ):
         if not cmd_path:
             return self
 
@@ -216,9 +285,16 @@ class PicklableArgumentParser(ArgumentParser):
         if subparsers.choices and cmdname in subparsers.choices:
             subparser = subparsers.choices[cmdname]
         else:
-            subparser = subparsers.add_parser(cmdname)
+            full_path = prefix + cmd_path
+            helpstr = subcommand_help.get(full_path) if subcommand_help else None
+            if helpstr:
+                subparser = subparsers.add_parser(cmdname, help=helpstr)
+            else:
+                subparser = subparsers.add_parser(cmdname)
 
-        return subparser.get_nested_subparser(*cmd_path[1:])
+        return subparser.get_nested_subparser(
+            *cmd_path[1:], prefix=prefix + cmd_path[:1], subcommand_help=subcommand_help
+        )
 
 
 class CommandLineInterface(PicklableArgumentParser, Logged):
@@ -281,15 +357,21 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         ignore_function_defaults: bool = False,
         typecheck: bool = False,
         output_handler: Opt[Callable] = None,
+        # error handling
+        exit_codes: Opt[Mapping[Type[Exception], int]] = None,
+        # documentation
+        subcommand_help: Opt[Mapping[Union[str, Tuple[str, ...]], str]] = None,
         # special features and flags
         use_multiprocessing: bool = False,
-        install_bash_completion: bool = False,
         use_verbose_flag: bool = False,
         use_quiet_flag: bool = False,
         use_execution_flag: Union[bool, str, Tuple[str, ...]] = False,
-        add_install_bash_completion_flag: Union[bool, str] = True,
         add_init_config_command: Union[bool, str, Tuple[str, ...]] = False,
         suppress_setup_warnings: bool = False,
+        # completion
+        install_bash_completion: bool = False,
+        extra_bash_completion_script: Opt[str] = None,
+        add_install_bash_completion_flag: Union[bool, str] = True,
         # source files
         source_file: Opt[str] = None,
         helper_files: Opt[List[str]] = None,
@@ -384,9 +466,41 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         :param output_handler: optional callable. Should take the return value of the invoked command/function and
             perform (usually) some IO action on it, such as saving it to disk. The return value is passed as the first
             argument and any further args will be supplied as keyword args parsed from the CLI or config.
+        :param exit_codes: optional mapping from exception classes to exit codes. If your application raises an
+            exception, the exit code will be equal to the highest code among those corresponding to the exception
+            classes matching the raised exception (where 'match' here means that the raised exception is an instance of
+            an exception class key in the exit_codes mapping but no other key is more specific).
+
+        :param subcommand_help: optional mapping from subcommand path (as space-separated string of, or tuple of,
+         command components) to help string, as would be accepted by argparse's .add_subparser(help=<help string>)
+         method. This allows for documentation of internal subcommands which serve only as parent command groups for
+         funtions or methods.
+
         :param use_multiprocessing: bool. If your app uses multiprocessing, then logging will be configured to reflect
             that fact, using appropriate process-safe loggers and handlers. This is passed through to
             `application.logging.config.configure_default_logging` via the `multiprocessing` keyword arg.
+        :param use_verbose_flag: bool. If passed, a flag is added to the command line interface using the option strings
+            `application.cli.VERBOSE_FLAGS` which may be repeated to increase verbosity. This affects verbosity by
+             decreasing the logging level by 10 for every repetition. At the DEBUG level (usually 3 repetitions),
+             the log format also changes to reflect more information, such as source files and line numbers.
+        :param use_quiet_flag: bool. If passed, a flag is added to the command line interface using the option strings
+            `application.cli.QUIET_FLAGS` with the effect that when the flag is passed, console logging is suppressed.
+        :param use_execution_flag: bool or str. When True or a str, a flag is added to the command line interface which
+            is interpreted as specifying that file-system-altering actions may be carried out, with the default behavior
+            being to skip these actions, possibly with verbose reporting as in a "dry-run" scenario. Your application
+            code may look up the status of this flag via `from bourbaki.application import cli; if cli.EXECUTE: ...`
+            to determine what action to take. The only effect on the behavior of this class is that file logging is
+            suppressed when the flag is not passed (when this arg is specified).
+        :param add_init_config_command: bool or str. When True or a str, a command is added to the command line
+            interface which wraps `application.CommandLineInterface.init_config`. This command writes an empty
+            configuration file (or dir) for your command line interface that can then be manually edited and passed
+            to the --config option when that option is available. See `application.CommandLineInterface.init_config` for
+            more details.
+        :param add_install_bash_completion_flag: bool or str. If True or a str, a flag is added to the command line
+            interface which triggers installation of bash completions when it is passed. If a str, that flag is equal to
+            this arg, else the default flag is `application.cli.INSTALL_SHELL_COMPLETION_FLAG`. Note that the
+            'bash-completion' package may need to be installed for your OS for some completions to work; see the
+            documentation for `application.completion` for more details.
         :param install_bash_completion: bool. If True, shell completions are installed automatically at the end of
             interface inference in a `CommandLineInterface.definition` decorator call, if the source file(s) have
             changed more recently than the completion definition files or the completion definition files don't yet
@@ -394,28 +508,8 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
             `CommandLineInterface.subcommand`, this option has no effect, since it is unknown when the CLI definition
             is complete. In that case, you can manually call `CommandLineInterface.install_shell_completion` in your
             script. Also see `add_install_bash_completion_flag`.
-        :param use_verbose_flag: bool. If passed, a flag is added to the command line interface using the option strings
-            `application.cli.VERBOSE_FLAGS` which may be repeated to increase verbosity. This affects verbosity by
-             decreasing the logging level by 10 for every repetition. At the DEBUG level (usually 4 repetitions),
-             the log format also changes to reflect more information, such as source files and line numbers.
-        :param use_quiet_flag: bool. If passed, a flag is added to the command line interface using the option strings
-            `application.cli.QUIET_FLAGS` with the effect that when the flag is passed, console logging is suppressed.
-        :param use_execution_flag: bool or str. When True or a str, a flag is added to the command line interface which
-            is interpreted as specifying that file-system-altering actions may be carried out, with the default behavior
-            being to skip these actions, possibly with verbose reporting as in a "dry-run" scenario. Your application
-            code may look up the status of this flag via `from application import cli; if cli.EXECUTE: ...` to determine
-            what action to take. The only effect that is determined by this class is that file logging is suppressed
-            when the flag is passed.
-        :param add_install_bash_completion_flag: bool or str. If True or a str, a flag is added to the command line
-            interface which triggers installation of bash completions when it is passed. If a str, that flag is equal to
-            this arg, else the default flag is `application.cli.INSTALL_SHELL_COMPLETION_FLAG`. Note that the
-            'bash-completion' package may need to be installed for your OS for some completions to work; see the
-            documentation for `application.completion` for more details.
-        :param add_init_config_command: bool or str. When True or a str, a command is added to the command line
-            interface which wraps `application.CommandLineInterface.init_config`. This command writes an empty
-            configuration file (or dir) for your command line interface that can then be manually edited and passed
-            to the --config option when that option is available. See `application.CommandLineInterface.init_config` for
-            more details.
+        :param extra_bash_completion_script: str. If given, append this script to the completion script for this CLI.
+            This allows for e.g. injection of user-defined shell functions to be used as completions.
 
         :param suppress_setup_warnings: bool. During processing of these args and inference of the interface from
             annotations and docstrings, some warnings may arise. To suppress these, pass True.
@@ -597,18 +691,18 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
                     "both package and version were passed explicitly; the given version will be used rather "
                     "than the version parsed from package info"
                 )
-                self.add_argument("--version", action="version", version=version)
+                self.add_argument(VERSION_FLAG, action="version", version=version)
             elif version or version is None:
                 if version is None:
                     version = True
                 self.add_argument(
-                    "--version",
+                    VERSION_FLAG,
                     action=PackageVersionAction,
                     package=package,
                     version=version,
                 )
             self.add_argument(
-                "--info",
+                INFO_FLAG,
                 action=InfoAction,
                 package=package,
                 version=version,
@@ -622,7 +716,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
                     )
                 )
             else:
-                self.add_argument("--version", action="version", version=version)
+                self.add_argument(VERSION_FLAG, action="version", version=version)
 
         if add_install_bash_completion_flag:
             flag = (
@@ -632,6 +726,20 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
             )
             self._add_argument(flag, action=InstallShellCompletionAction)
 
+        if subcommand_help is not None:
+            if not isinstance(subcommand_help, Mapping):
+                raise TypeError(
+                    "subcommand_help must be a mapping from command path to help string; got {}".format(
+                        type(subcommand_help)
+                    )
+                )
+            self.subcommand_help = {
+                tuple(k.strip().split() if isinstance(k, str) else k): v
+                for k, v in subcommand_help.items()
+            }
+        else:
+            self.subcommand_help = None
+
         self.version = version
         self.package = package
 
@@ -655,6 +763,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         self.parse_config_as_cli = parse_config_as_cli
         self.typecheck = typecheck
         self.output_handler = output_handler
+        self.exit_codes = exit_codes or {Exception: 1}
         self.require_options = bool(require_options)
 
         self.use_multiprocessing = bool(use_multiprocessing)
@@ -672,6 +781,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         else:
             self.source_dir = None
         self.helper_files = helper_files
+        self.extra_bash_completion_script = extra_bash_completion_script
         self._bash_completion = bool(install_bash_completion)
         self.use_init_config_command = bool(add_init_config_command)
         self.subcommands = {}
@@ -710,7 +820,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
                 self.init_config_command = tuple(self.init_config.__name__.split("_"))
             else:
                 if isinstance(add_init_config_command, str):
-                    add_init_config_command = (add_init_config_command,)
+                    add_init_config_command = add_init_config_command.strip().split()
 
                 self.init_config_command = tuple(
                     map(to_cmd_line_name, add_init_config_command)
@@ -742,21 +852,19 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
                 for a in subcommand.parser._actions
                 if a.dest in ("only_commands", "omit_commands")
             ]
+            all_command_choices = filter(
+                bool,
+                self.all_subcommand_names(
+                    include_prefixes=True,
+                    filter_=lambda cmd: bool(cmd.config_subsections),
+                ),
+            )
+            all_command_choices_str = list(map(" ".join, all_command_choices))
             for arg in commands_args:
                 if arg.choices is None:
                     arg.choices = set()
-                arg.choices.update(
-                    map(
-                        " ".join,
-                        filter(
-                            bool,
-                            self.all_subcommand_names(
-                                include_prefixes=True,
-                                filter_=lambda cmd: bool(cmd.config_subsections),
-                            ),
-                        ),
-                    )
-                )
+
+                arg.choices.update(all_command_choices_str)
                 arg.metavar = "CMD-NAME"
                 choices_str = (
                     "{{{}}}".format(",".join(map(shlex.quote, sorted(arg.choices))))
@@ -815,6 +923,22 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
     def cmd_name(self):
         return self.prog or os.path.split(sys.argv[0])[-1]
 
+    ##############
+    # subparsers #
+    ##############
+
+    def get_nested_subparser(
+        self,
+        *cmd_path: str,
+        prefix: Tuple[str, ...] = (),
+        subcommand_help: Opt[Mapping[Tuple[str, ...], str]] = None,
+    ):
+        if subcommand_help is None:
+            subcommand_help = self.subcommand_help
+        return super().get_nested_subparser(
+            *cmd_path, prefix=prefix, subcommand_help=subcommand_help
+        )
+
     def all_subcommands(
         self,
         *prefix: str,
@@ -869,7 +993,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         self,
         args=None,
         namespace=None,
-        report_progress=True,
+        report_progress=False,
         time_units="s",
         log_level=PROGRESS,
         error_level=ERROR,
@@ -891,32 +1015,41 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         else:
             main = False
 
-        app_logger = self.get_app_logger(ns)
-        app_logger.debug("command is %r", cmdname)
-        config = self.parse_config(ns, app_logger) if self.use_config else None
+        verbosity = getattr(ns, VERBOSITY_ATTR, MIN_VERBOSITY)
+        if cmdfunc.exit_codes is None:
+            exit_codes = self.exit_codes
+        else:
+            exit_codes = ChainMap(cmdfunc.exit_codes, self.exit_codes)
 
-        # if self is defined from a class and the command is not a reserved/builtin command,
-        if (
-            (not main)
-            and (self.app_cls is not None)
-            and (cmdname not in self.reserved_command_names)
+        with CLIErrorHandlingContext(
+            exit_codes, verbose=verbosity >= TRACEBACK_VERBOSITY
         ):
-            # perform any initialization logic; if main == True, this will be done below at func.execute()
-            app_obj = self._main.execute(ns, config, handle_output=False)
-        else:
-            app_obj = None
+            app_logger = self.get_app_logger(ns)
+            app_logger.debug("command is %r", cmdname)
+            config = self.parse_config(ns, app_logger) if self.use_config else None
 
-        if not report_progress:
-            result = cmdfunc.execute(ns, config, app_obj)
-        else:
-            with get_task(
-                app_logger,
-                cmdname,
-                log_level=log_level,
-                error_level=error_level,
-                time_units=time_units,
+            # if self is defined from a class and the command is not a reserved/builtin command,
+            if (
+                (not main)
+                and (self.app_cls is not None)
+                and (cmdname not in self.reserved_command_names)
             ):
+                # perform any initialization logic; if main == True, this will be done below at func.execute()
+                app_obj = self._main.execute(ns, config, handle_output=False)
+            else:
+                app_obj = None
+
+            if not report_progress:
                 result = cmdfunc.execute(ns, config, app_obj)
+            else:
+                with get_task(
+                    app_logger,
+                    cmdname,
+                    log_level=log_level,
+                    error_level=error_level,
+                    time_units=time_units,
+                ):
+                    result = cmdfunc.execute(ns, config, app_obj)
 
         return result
 
@@ -1075,6 +1208,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
     def subcommand(
         self,
         command_prefix=None,
+        name=None,
         config_subsections=None,
         implicit_flags=None,
         ignore_on_cmd_line=None,
@@ -1084,9 +1218,9 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         parse_order=None,
         typecheck=None,
         output_handler=None,
+        exit_codes=None,
         named_groups=None,
         require_options=None,
-        name=None,
         from_method=False,
         metavars=None,
         tvar_map=None,
@@ -1101,15 +1235,28 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
 
         def dec(
             f,
+            # have to pass these in so we can rebind locals if needed below
             command_prefix=command_prefix,
             output_handler=output_handler,
+            exit_codes=exit_codes,
             config_subsections=config_subsections,
+            tvar_map=tvar_map,
         ):
             # args provided here override those specified with decorators on the function, which override this
             # command line interface's global defaults
             if output_handler is None:
                 output_handler = cli_attrs.output_handler(f, None)
 
+            if exit_codes is None:
+                exit_codes = cli_attrs.exit_codes(f)
+
+            if tvar_map is None:
+                try:
+                    # for callable instances of generic classes
+                    tvar_map = get_param_dict(get_generic_type(f))
+                except (AttributeError, TypeError):
+                    pass
+
             if output_handler is None:
                 if _main and self.require_subcommand:
                     output_handler = NO_OUTPUT_HANDLER
@@ -1154,6 +1301,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
                 command_prefix=command_prefix,
                 config_subsections=config_subsections,
                 output_handler=output_handler,
+                exit_codes=exit_codes,
                 implicit_flags=implicit_flags,
                 lookup_order=self.lookup_order,
                 argparser_cmd_name=self.cmd_name,
@@ -1189,6 +1337,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
         parse_order=None,
         typecheck=None,
         output_handler=None,
+        exit_codes=None,
         named_groups=None,
         name=None,
         from_method=False,
@@ -1204,6 +1353,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
             parse_config_as_cli=parse_config_as_cli,
             typecheck=typecheck,
             output_handler=output_handler,
+            exit_codes=exit_codes,
             named_groups=named_groups,
             parse_order=parse_order,
             metavars=metavars,
@@ -1261,7 +1411,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
                         )
                     )
                 continue
-            
+
             cmd_name = cli_attrs.command_name(f, default=name)
             self.subcommand(name=cmd_name, from_method=True, tvar_map=tvar_map)(f)
 
@@ -1291,7 +1441,7 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
 
             self._set_subcommand(*cmd_path, subcommand=subcmd_func)
         else:
-            parser = self.add_argument_group("main arguments")
+            parser = self
 
             if self.description is None:
                 self.description = subcmd_func.docs.short_desc
@@ -1364,7 +1514,12 @@ class CommandLineInterface(PicklableArgumentParser, Logged):
             names = (self.prog, os.path.basename(self.source_file))
 
         self.logger.debug("Installing bash completion for command `{}`".format(names))
-        install_shell_completion(self, *names, last_edit_time=self.last_edit_time())
+        install_shell_completion(
+            self,
+            *names,
+            extra_completion_script=self.extra_bash_completion_script,
+            last_edit_time=self.last_edit_time(),
+        )
 
     ##################
     # pickle methods #
@@ -1556,6 +1711,7 @@ class SubCommandFunc(Logged):
         config_subsections=None,
         output_handler=None,
         implicit_flags=False,
+        exit_codes=None,
         argparser_cmd_name=None,
         suppress_setup_warnings=False,
         tvar_map=None,
@@ -1701,6 +1857,7 @@ class SubCommandFunc(Logged):
         self.output_docs = output_docs
         self.func = func
         self.output_handler = output_handler
+        self.exit_codes = exit_codes
         self.from_method = bool(from_method)
 
         # options
@@ -2105,12 +2262,37 @@ class SubCommandFunc(Logged):
                     elif not isinstance(val, list):  # tuples
                         val = list(val)
             elif literal_defaults and has_nonnull_default:
+                # encode the value
                 val = tio.config_encoder(param.default)
             else:
-                val = tio.config_repr
+                # represent the type
+                # at this point we either have literal_defaults=False or has_nonnull_default=False
+                # in either case we would want a type repr
+                is_optional = is_optional_type(tio.type_)
+                if is_optional and has_default:
+                    # indicate the key can be ommitted by specially formatting it
+                    name = OPTIONAL_ARG_TEMPLATE.format(name)
+                    if has_nonnull_default:
+                        # include the null option in the type repr to indicate that a null override is possible
+                        val = tio.config_repr
+                    else:
+                        # only repr the non-null type options when there is a default, to simplify
+                        NoneType = type(None)
+                        types = tuple(
+                            t for t in get_generic_args(tio.type_) if t is not NoneType
+                        )
+                        if len(types) == 1:
+                            val = TypedIO(types[0]).config_repr
+                        else:
+                            val = TypedIO(Union[types]).config_repr
+                else:
+                    # non-option type with a default, or option type with no default;
+                    # in either case we want the plain type repr
+                    val = tio.config_repr
 
-            if is_optional_type(tio.type_) and has_default:
-                name = OPTIONAL_ARG_TEMPLATE.format(name)
+                # deepcopy here prevents config formats with references (e.g. yaml) from using that feature,
+                # which results in confusing output
+                val = copy.deepcopy(val)
 
             conf[name] = val
 
